@@ -8,10 +8,13 @@ use anyhow::{anyhow, Context, Result};
 use futures_util::Future;
 use nix::sys::socket::{self as nixsocket, ControlMessageOwned};
 use nix::sys::uio::IoVec;
+use once_cell::sync::Lazy;
 use serde::{Deserialize, Serialize};
+use std::convert::TryFrom;
 use std::fs::File;
 use std::os::unix::io::AsRawFd;
 use std::os::unix::prelude::{CommandExt, FromRawFd, RawFd};
+use std::path::PathBuf;
 use std::pin::Pin;
 use std::process::{Command, Stdio};
 use std::sync::{Arc, Mutex};
@@ -27,11 +30,12 @@ pub const OCI_TYPE_LAYER_TAR: &str = "application/vnd.oci.image.layer.v1.tar";
 // Note that payload data (non-metadata) should go over a pipe file descriptor.
 const MAX_MSG_SIZE: usize = 32 * 1024;
 
-lazy_static::lazy_static! {
-    static ref SUPPORTED_PROTO_VERSION: semver::VersionReq = {
-        semver::VersionReq::parse("0.2.0").unwrap()
-    };
-}
+static BASE_PROTO_VERSION: Lazy<semver::VersionReq> =
+    Lazy::new(|| semver::VersionReq::parse("0.2.0").unwrap());
+
+// https://github.com/containers/skopeo/pull/1523
+static SEMVER_0_2_3: Lazy<semver::VersionReq> =
+    Lazy::new(|| semver::VersionReq::parse("0.2.3").unwrap());
 
 #[derive(Serialize)]
 struct Request {
@@ -76,6 +80,7 @@ type ChildFuture = Pin<
 pub struct ImageProxy {
     sockfd: Arc<Mutex<File>>,
     childwait: Arc<AsyncMutex<ChildFuture>>,
+    protover: semver::Version,
 }
 
 impl std::fmt::Debug for ImageProxy {
@@ -116,9 +121,20 @@ fn file_from_scm_rights(cmsg: ControlMessageOwned) -> Option<File> {
 #[derive(Debug, Default)]
 pub struct ImageProxyConfig {
     /// Path to container auth file; equivalent to `skopeo --authfile`.
-    pub authfile: Option<String>,
+    pub authfile: Option<PathBuf>,
+
+    /// Do not use default container authentication paths; equivalent to `skopeo --no-creds`.
+    ///
+    /// Defaults to `false`; in other words, use the default file paths from `man containers-auth.json`.
+    pub auth_anonymous: bool,
+
+    // Directory with certificates (*.crt, *.cert, *.key) used to connect to registry
+    // Equivalent to `skopeo --cert-dir`
+    pub certificate_directory: Option<PathBuf>,
+
     /// If set, disable TLS verification.  Equivalent to `skopeo --tls-verify=false`.
     pub insecure_skip_tls_verification: Option<bool>,
+
     /// Provide a configured [`std::process::Command`] instance.
     ///
     /// This allows configuring aspects of the resulting child `skopeo` process.
@@ -142,16 +158,10 @@ pub struct ImageProxyConfig {
     pub skopeo_cmd: Option<Command>,
 }
 
-impl ImageProxy {
-    /// Create an image proxy that fetches the target image, using default configuration.
-    pub async fn new() -> Result<Self> {
-        Self::new_with_config(Default::default()).await
-    }
+impl TryFrom<ImageProxyConfig> for Command {
+    type Error = anyhow::Error;
 
-    /// Create an image proxy that fetches the target image
-    #[instrument]
-    pub async fn new_with_config(config: ImageProxyConfig) -> Result<Self> {
-        let (mysock, theirsock) = new_seqpacket_pair()?;
+    fn try_from(config: ImageProxyConfig) -> Result<Self, Self::Error> {
         // By default, we set up pdeathsig to "lifecycle bind" the child process to us.
         let mut c = config.skopeo_cmd.unwrap_or_else(|| {
             let mut c = std::process::Command::new("skopeo");
@@ -164,13 +174,39 @@ impl ImageProxy {
             c
         });
         c.arg("experimental-image-proxy");
-        if let Some(authfile) = config.authfile.as_deref() {
-            c.args(&["--authfile", authfile]);
+        if let Some(authfile) = config.authfile {
+            if config.auth_anonymous {
+                // This is a programmer error really
+                anyhow::bail!("Cannot use anonymous auth and an authfile");
+            }
+            c.arg("--authfile");
+            c.arg(authfile);
+        } else if config.auth_anonymous {
+            c.arg("--no-creds");
+        }
+        if let Some(certificate_directory) = config.certificate_directory {
+            c.arg("--cert-dir");
+            c.arg(certificate_directory);
         }
         if config.insecure_skip_tls_verification.unwrap_or_default() {
             c.arg("--tls-verify=false");
         }
         c.stdout(Stdio::null()).stderr(Stdio::piped());
+        Ok(c)
+    }
+}
+
+impl ImageProxy {
+    /// Create an image proxy that fetches the target image, using default configuration.
+    pub async fn new() -> Result<Self> {
+        Self::new_with_config(Default::default()).await
+    }
+
+    /// Create an image proxy that fetches the target image
+    #[instrument]
+    pub async fn new_with_config(config: ImageProxyConfig) -> Result<Self> {
+        let mut c = Command::try_from(config)?;
+        let (mysock, theirsock) = new_seqpacket_pair()?;
         c.stdin(Stdio::from(theirsock));
         let child = c.spawn().context("Failed to spawn skopeo")?;
         tracing::debug!("Spawned skopeo pid={:?}", child.id());
@@ -181,15 +217,20 @@ impl ImageProxy {
         let childwait = tokio::task::spawn_blocking(move || child.wait_with_output());
         let sockfd = Arc::new(Mutex::new(mysock));
 
-        let r = Self {
+        let mut r = Self {
             sockfd,
             childwait: Arc::new(AsyncMutex::new(Box::pin(childwait))),
+            protover: semver::Version::new(0, 0, 0),
         };
 
         // Verify semantic version
         let protover = r.impl_request::<String, _, ()>("Initialize", []).await?.0;
         let protover = semver::Version::parse(protover.as_str())?;
-        let supported = &*SUPPORTED_PROTO_VERSION;
+        let supported = if cfg!(feature = "proxy_v0_2_3") {
+            &*SEMVER_0_2_3
+        } else {
+            &*BASE_PROTO_VERSION
+        };
         if !supported.matches(&protover) {
             return Err(anyhow!(
                 "Unsupported protocol version {} (compatible: {})",
@@ -197,6 +238,7 @@ impl ImageProxy {
                 supported
             ));
         }
+        r.protover = protover;
 
         Ok(r)
     }
@@ -305,18 +347,32 @@ impl ImageProxy {
         Ok(r)
     }
 
+    async fn read_all_fd(&self, fd: Option<(File, u32)>) -> Result<Vec<u8>> {
+        let (fd, pipeid) = fd.ok_or_else(|| anyhow!("Missing fd from reply"))?;
+        let mut fd = tokio::io::BufReader::new(tokio::fs::File::from_std(fd));
+        let mut r = Vec::new();
+        let reader = fd.read_to_end(&mut r);
+        let (nbytes, finish) = tokio::join!(reader, self.finish_pipe(pipeid));
+        finish?;
+        assert_eq!(nbytes?, r.len());
+        Ok(r)
+    }
+
     /// Fetch the manifest.
     /// For more information on OCI manifests, see <https://github.com/opencontainers/image-spec/blob/main/manifest.md>
     pub async fn fetch_manifest(&self, img: &OpenedImage) -> Result<(String, Vec<u8>)> {
         let (digest, fd) = self.impl_request("GetManifest", [img.0]).await?;
-        let (fd, pipeid) = fd.ok_or_else(|| anyhow!("Missing fd from reply"))?;
-        let mut fd = tokio::io::BufReader::new(tokio::fs::File::from_std(fd));
-        let mut manifest = Vec::new();
-        let reader = fd.read_to_end(&mut manifest);
-        let (reader, finish) = tokio::join!(reader, self.finish_pipe(pipeid));
-        reader?;
-        finish?;
-        Ok((digest, manifest))
+        Ok((digest, self.read_all_fd(fd).await?))
+    }
+
+    /// Fetch the config.
+    /// For more information on OCI config, see <https://github.com/opencontainers/image-spec/blob/main/config.md>
+    #[cfg(feature = "proxy_v0_2_3")]
+    pub async fn fetch_config(&self, img: &OpenedImage) -> Result<Vec<u8>> {
+        let (_, fd) = self
+            .impl_request::<(), _, _>("GetFullConfig", [img.0])
+            .await?;
+        self.read_all_fd(fd).await
     }
 
     /// Fetch a blob identified by e.g. `sha256:<digest>`.
@@ -365,3 +421,52 @@ impl ImageProxy {
         Ok(())
     }
 }
+
+#[cfg(test)]
+mod tests {
+    use super::*;
+
+    fn validate(c: Command, contains: &[&str], not_contains: &[&str]) {
+        // Format via debug, because
+        // https://doc.rust-lang.org/std/process/struct.Command.html#method.get_args
+        // is experimental
+        let d = format!("{:?}", c);
+        for c in contains {
+            assert!(d.contains(c), "{} missing {}", d, c);
+        }
+        for c in not_contains {
+            assert!(!d.contains(c), "{} should not contain {}", d, c);
+        }
+    }
+
+    #[test]
+    fn proxy_configs() {
+        let c = Command::try_from(ImageProxyConfig::default()).unwrap();
+        validate(
+            c,
+            &["experimental-image-proxy"],
+            &["--no-creds", "--tls-verify", "--authfile"],
+        );
+
+        let c = Command::try_from(ImageProxyConfig {
+            authfile: Some(PathBuf::from("/path/to/authfile")),
+            ..Default::default()
+        })
+        .unwrap();
+        validate(c, &[r"--authfile", "/path/to/authfile"], &[]);
+
+        let c = Command::try_from(ImageProxyConfig {
+            certificate_directory: Some(PathBuf::from("/path/to/certs")),
+            ..Default::default()
+        })
+        .unwrap();
+        validate(c, &[r"--cert-dir", "/path/to/certs"], &[]);
+
+        let c = Command::try_from(ImageProxyConfig {
+            insecure_skip_tls_verification: Some(true),
+            ..Default::default()
+        })
+        .unwrap();
+        validate(c, &[r"--tls-verify=false"], &[]);
+    }
+}
